/*******************************************************************************
* File Name    : LCD_LIB
* Version      : Rev. 2.00
* Device(s)    : R7F100GL
* Tool-Chain   : Arduino
* Description  : This file implements LCD control function for ACM1602
* Creation Date: 2016/09/30 First Release
* Modified Date: 2022/05/20 Ported to RL78/G23-64PFPB
*                2023/04/25 Ported to RL78/G22-48PFPB
********************************************************************************/
#include <Wire.h>
#include <Wire_IICA.h>
#include "LCD_LIB.h"

#define SEL_COMMAND         (0U)        // parameter for RW pin to set command 
#define SEL_DATA            (1U)        // parameter for RW pin to set data

/* Maximum characters per line of LCD display. */
#define LCD_SIZE            16
/* number of lines on the LCD display */
#define MAXIMUM_LINES       2

/* LCD data address for display (MAX 40 charactor/line)  */
#define LCD_LINE1 0x00                  /* first line start from address 0x00   */
#define LCD_LINE2 0x40                  /* 2nd line start from address 0x40     */


/*****************************************************************************
LCD commands - use LCD_write function to write these commands to the LCD. 
*****************************************************************************/
/* Clear LCD display (set 20H) and home cursor              wait 1.64ms */
#define LCD_CLEAR        0b00000001
/*                                +--:clear display command */

/* Cursor auto decrement after R/W (not shift display)      wait 40us   */
#define CURSOR_MODE_DEC  0b00000100

/* Cursor auto increment after R/W (not shift display)  */
#define CURSOR_MODE_INC  0b00000110
/*                              ||+--:SH:shift display position
                                ||  * 0:not shift
                                ||  * 1:shift display
                                |+---:I/D:Incement/Decrement pointer
                                |     0:move pointer to left 
                                |   * 1:move pointer to right
                                +----:entry mode set command    */

/* Display ON with Cursor                                   wait 40us   */
#define LCD_CURSOR_ON    0b00001110
/* Display ON with Cursor off   */
#define LCD_CURSOR_OFF   0b00001100
/* Display on with blinking cursor  */
#define LCD_CURSOR_BLINK 0b00001101
/*                             |||+--:B:Cursor blink
                               |||  * 0:blink off
                               |||  * 1:blink on
                               ||+---:C:Cursor ON/OFF
                               ||   * 0:disappered
                               ||   * 1:cursor is turned on
                               |+----:D:Display ON/OFF
                               |      0:display is turned off
                               |    * 1:display is turned on
                               +-----Display ON/OFF command */

/*Move Cursor Left One Position                             wait 40us   */
#define LCD_CURSOR_LEFT  0b00010000
/* Move Cursor Right One Position */
#define LCD_CURSOR_RIGHT 0b00010100
/* Move Display Right One Position */
#define LCD_DISP_LEFT    0b00011000
/* Move Display Right One Position */
#define LCD_DISP_RIGHT   0b00011100
/*                            |++----:S/C:R/L:Cursor or display shift mode
                              |    * 00:shift cursor to the left
                              |    * 01:shift cursor to the right
                              |      10:shift display and cursor to the left
                              |      11:shift display and cursor to the right
                              +-----Cursor or display shift command */

/* Setup, 4 bits bus,2 line display, 5x7 font               wait 40us   */
#define FUNCTION_SET     0b00101000
/*                           |||+----:F:Font type
                             |||    * 0:5*8 dot font
                             |||      1:5*11 dot font
                             ||+-----:N:Display line number
                             ||       0:1-line mode
                             ||     * 1:2-line mode
                             |+------:DL:Interface data length
                             |      * 0:4bit(D7-D4)bus mode
                             |        1:8bit(D7-D0)bus mode
                             +-------:Function set command  */

/* 1st nibble for 4bit I/F mode     */
#define F_4bitIF         0x02U
/* 2nd nibble for 4bit I/F mode     */
#define LCD_TWO_LINE     0x08U

/* move cursor to top of line 1 (data pointer)  */
#define LCD_HOME_L1      0b10000000
/*                         |+++++++--:data pointer(0x00:top of line1)
                           +---------:pointer set command   */
/* move cursor to top of line 2 (data pointer)  */
#define LCD_HOME_L2      0b11000000
/*                         |+++++++--:data pointer(0x40:top of line2)
                           +---------:pointer set command   */




volatile unsigned char slave7 = 0x3F;       // スレーブ・アドレス(7bit)

volatile uint8_t write_mode;                // PCF8574Aに書き込むデータのワーク(Rev3)
/*
    使用するシールドのPCF8574Aの結線に合わせ、以下のビット構成とする
    bit7:0  DB7 of HD44780
    bit6:0  DB6 of   〃
    bit5:0  DB5 of   〃
    bit4:0  DB4 of   〃
    bit3:1  バックライト制御(通常は1：点灯に)
    bit2:0  E信号 of HD44780
    bit1:0  R/W信号 of  〃      0:書き込み／1:読み出し
    bit0:0  RS信号 of   〃      0:コマンド／1:データ
*/

/*----------------------------------------------
    prototype declaration
------------------------------------------------*/
void set_2digit( uint8_t data);             // 引数をASCIIコード2桁で表示
void set_1digit( uint8_t data);             // 引数の下位4ビットをASCIIコード1桁で表示
byte LCD_write(uint8_t value);              // 引数をHD44780に書き込む
byte LCD_nibble_write(uint8_t value);       // 引数の下位4ビットをLCDに設定（コマンドorデータ）
byte PCF_Init(void);                        // PCF8574を初期化

/*******************************************************************************
LCD display control functions for G22_FPB
********************************************************************************/

/*******************************************************************************
* Function Name: init_LCD
* Description  : この関数で、LCDCを4ビットI/F、2行表示に設定
* Arguments    : none
* Return Value : none
********************************************************************************/
void init_LCD_S(void)
{

    PCF_Init();                             // PCF8574のEN信号をローにする

    write_mode = 0x08;                      // コマンド書き込みに設定しておく

    delay(100);                             // 電源電圧が2.7Vに達してから100ms以上待つ 

/*==================================================
    ここからが実際の初期化処理
    initialise LCDC(HD44780 or equivalent)
==================================================*/
/*----------------------------------------------
    set 4bit interface mode(preprocessing)
    LCDが4/8bitI/Fモードでも一旦8bitにしてから4bitにする
    （RL78だけがリセットされたときの対策）
    50マイクロ秒の待ち合わせはI2Cバス経由では不要だが、
    そのまま残しておきます。
------------------------------------------------*/

    LCD_nibble_write(0x03);                 // ダミーで8ビット幅に設定
    delay (5);                              // 5ミリ秒待つ
    LCD_nibble_write(0x03);                 // ダミー
    delayMicroseconds(100);                 // 100マイクロ秒待つ
    LCD_nibble_write(0x03);                 // ダミー
    delayMicroseconds(50);                  // 50マイクロ秒待つ

/*----------------------------------------------
    set  high nibble of 4bit I/F command only
    I/Fだけを4ビットに変更（下位4ビットはまだ不定）
------------------------------------------------*/

    LCD_nibble_write(F_4bitIF);             // 4ビット幅に変更
    delayMicroseconds(50);                  // 50マイクロ秒待つ
    
/*----------------------------------------------
    set complete 4bit I/F command
    ここからは，4ビットI/FでI/F条件を正しく設定する
------------------------------------------------*/

    LCD_nibble_write(F_4bitIF);             // LCDの指定で，再度設定
    LCD_nibble_write(LCD_TWO_LINE);         // 表示は2行
    delayMicroseconds(50);                  // 50マイクロ秒待つ（10マイクロでも十分）

/*----------------------------------------------
    表示状態の設定
------------------------------------------------*/

    LCD_write(LCD_CURSOR_OFF);              // カーソル非表示でLCD表示
    delayMicroseconds(50);                  // 50マイクロ秒待つ（10マイクロでも十分）

/*----------------------------------------------
    LCD表示をクリアする
------------------------------------------------*/

    LCD_write(LCD_CLEAR);                   // LCDのクリアコマンド送信
    delayMicroseconds (2100);               // クリア処理時間待ち(2.1ms) 

/*----------------------------------------------
    データ入力モード設定
------------------------------------------------*/

    LCD_write(CURSOR_MODE_INC);             // 書き込みで右に移動
    delayMicroseconds(50);                  // 50マイクロ秒待つ（10マイクロでも十分）
    LCD_write(LCD_HOME_L1);                 // 1行目の1文字目に移動 
    delayMicroseconds(50);                  // 50マイクロ秒待つ（10マイクロでも十分）

}
/******************************************************************************
End of function init_LCD_S function
******************************************************************************/

/*******************************************************************************
* Function Name:  PCF_Init
* Description  : この関数で、PCF8574のbit2（E信号）をネゲートする
* Arguments    : none
* Return value  : I2Cのステータス
*                   00: 成功
*                   01: 送信バッファのオーバーフロー
*                   02: スレーブ・アドレスにNACK応答
*                   03: 送信データにNACK応答
*                   04: その他のエラー
********************************************************************************/
byte PCF_Init(void)
{

    Wire.beginTransmission( slave7 );   // PCF8574のアドレス指定
    Wire.write(0xFB);                   // E信号をネゲートするデータ

    return( Wire.endTransmission(1));   // 実際に送信する

}
/******************************************************************************
End of function PCF_Init function
******************************************************************************/


/*******************************************************************************
* Function Name: setCursor
* Description  : この関数でカーソル位置を指定する
* Arguments    : cursor address
*               col: 0 to 0x0F
*               row : 0(line1) /1(line2)
* Return Value : none
********************************************************************************/
void setCursor( uint8_t col , uint8_t row)
{

    uint8_t work;

    work = (( row & 1 ) * 0x40);            // 行の先頭アドレス作成
    work += ( col & 0x0F );                 // カーソル位置を作成
    work |= 0x80;                           // カーソル移動コマンド作成

    LCD_write( work );                      // カーソルを移動
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function setCursor
******************************************************************************/


/*******************************************************************************
* Function Name: set_2digitBCD
* Description  : この関数は引数（2進数の99-0）を十進数としてASCII変換して、2桁で表示
* Arguments    : display number(99-0)
* Return Value : none
********************************************************************************/
void set_2digitBCD( uint8_t data)
{

    uint8_t work;

    work = data;                            // 表示データをセット
    
    set_1digit((work / 10));                // 10の桁を表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ
    set_1digit((work % 10));                // 1の桁を表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function set_2digitBCD
******************************************************************************/


/*******************************************************************************
* Function Name: set_2digit
* Description  : この関数は引数（8ビット）をASCII変換して2桁で表示
* Arguments    : display number(BCD or HEX)
* Return Value : none
********************************************************************************/
void set_2digit( uint8_t data)
{

    uint8_t work;

    work = data;                            // 表示データをセット

    set_1digit( work >> 4 );                // 上位4bitを表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ
    set_1digit( data );                     // 下位4bitを表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function set_2digit
******************************************************************************/


/*******************************************************************************
* Function Name: set_1digit
* Description  : この関数は引数の下位4ビットをASCII変換して表示
* Arguments    : display lowwer 4bit as ASCII
* Return Value : none
********************************************************************************/
void set_1digit( uint8_t data)
{

    uint8_t work;

    work = (data & 0x0F);                   // 未使用ビットをマスク
    if ( work < 0x0A )
    {
        work += 0x30;                       // 0-9をASCIIコードに変換
    }
    else
    {
        work += 0x37;                       // A-FをASCIIコードに変換
    }

    LCD_DATA_write(work);                   // 1桁のデータ（0-F）を表示

}
/******************************************************************************
End of function set_1digit
******************************************************************************/


/*******************************************************************************
* Function name     : DisplayString
* Description       : この関数はLCDの第1引数で示す位置から第2引数で示すデータを表示する
*                     第1引数は0-63なら1行目、64-127なら2行目を示す
*                     表示データはASCIIコードとし、0x00で終端する
*                     For example, to start at the 2nd position on line 1...
*                     DisplayString(LCD_LINE1 + 1, string address) 
* Argument          : (uint8_t) LCD position 
*                     (int8_t *) string
* Return value      : none.
* Remarks           : 表示用RAM領域は2行なら 1行目:0-27H(=39）,2行目:40H(=64)-67H(103)
*
*                     4行表示LCDでは、1行目:0-13H(19)、2行目:40H(=64)-53H(83)、
*                                     3行目:14H(20)-27H(39)、4行目:54H(84)-67H(103)
*******************************************************************************/
void DisplayString(uint8_t position, int8_t * string)
{

    if ( position < LCD_LINE2 )
    {                                         // カーソルをline1に移動させる
        /* Display on Line 1 */
        LCD_write((uint8_t)(LCD_HOME_L1 + position));
    }
    else
    {                                         // カーソルをline2に移動させる
        /* Display on Line 2 */
        LCD_write((uint8_t)(LCD_HOME_L2 + (position - LCD_LINE2)));
    }

    delayMicroseconds(60);                    // 60マイクロ秒待つ
    write_mode = 0x09;                        // データを選択

    do
    {
        LCD_write(*string++);                 // 表示データ転送
        delayMicroseconds(60);                // 60マイクロ秒待つ

    }
    while (*string);

    write_mode = 0x08;                        // コマンド転送に戻す

}
/******************************************************************************
End of function DisplayString
******************************************************************************/


/******************************************************************************
* Function name : LCD_DATA_write
* Description   : この関数は引数を表示データとしてLCDに書き込む
* Argument      : (uint8_t)value - 表示データ
* Return value  : I2Cのステータス
*                   00: 成功
*                   01: 送信バッファのオーバーフロー
*                   02: スレーブ・アドレスにNACK応答
*                   03: 送信データにNACK応答
*                   04: その他のエラー
* Remarks       : 変数slave7にスレーブアドレスが格納されているものとする
*******************************************************************************/
byte LCD_DATA_write(uint8_t value)
{

    byte result;                            // 戻り値用

    write_mode = 0x09;                      // bit0=1で表示データを選択

    result = LCD_write(value);              // 表示データ転送

    write_mode = 0x08;                      // コマンド（bit0=0）を選択

    return(result);                         // 通信結果を戻す

}
/******************************************************************************
End of function LCD_DATA_write
******************************************************************************/

/******************************************************************************
* Function name : LCD_write
* Description   : この関数は引数をコマンドとして書き込む
* Argument      : (uint8_t)value 書き込みデータ／コマンド
* Return value  : I2Cのステータス
*                   00: 成功
*                   01: 送信バッファのオーバーフロー
*                   02: スレーブ・アドレスにNACK応答
*                   03: 送信データにNACK応答
*                   04: その他のエラー
* Remarks       : 変数write_modeのビット0にRS信号の状態が設定されているものとする
*                 変数slave7にスレーブアドレスが格納されているものとする
*******************************************************************************/
byte LCD_write(uint8_t value)
{

    uint8_t work;                           // データ書き込み作業用

    Wire.beginTransmission( slave7 );       // PCF8574のアドレス指定

/*----------------------------------------------
    上位ニブルを書き込み
------------------------------------------------*/

    work    = (write_mode & 0x09);          // RS信号をコピー,R/WをW(0)に,BLは1
    Wire.write(work);                       // 書き込み信号セット
    work    |= 0x04;                        // E信号セット
    work    |= (value & 0xF0);              // 書き込みデータ上位ビットをマージ
    Wire.write(work);                       // 書き込みデータとE信号アソート

    work    ^= 0x04;                        // E信号をクリア
    Wire.write(work);                       // E信号をネゲート

/*----------------------------------------------
    下位ニブルを書き込み
------------------------------------------------*/

    work    &= 0x0F;                        // データ領域をクリア
    work    |= 0x04;                        // E信号セット
    work    |= (value << 4);                // 書き込みデータを上位4bitに設定
    Wire.write(work);                       // 書き込みデータとE信号アソート

    work    ^= 0x04;                        // E信号をクリア
    Wire.write(work);                       // E信号をネゲート

    return(Wire.endTransmission(1));        // データ送信とストップコンディション完了

}
/******************************************************************************
End of function LCD_write
******************************************************************************/


/*******************************************************************************
* Function name : LCD_write_nibble
* Description   : この関数は引数の下位4ビットをLCDに書き込む
* Argument      : (uint8_t) value - the value to write(low nibble)
* Return value  : I2Cのステータス
*                   00: 成功
*                   01: 送信バッファのオーバーフロー
*                   02: スレーブ・アドレスにNACK応答
*                   03: 送信データにNACK応答
*                   04: その他のエラー
* Remarks       : 変数write_modeのビット0にRS信号の状態が設定されているものとする
*                 変数slave7にスレーブアドレスが格納されているものとする
*******************************************************************************/
byte LCD_nibble_write(uint8_t value)
{
    uint8_t work;                           // データ書き込み作業用

    Wire.beginTransmission( slave7 );       // PCF8574のアドレス指定

    work    = (write_mode & 0b00001001);    // RS信号をコピー,R/WをWに
    Wire.write(work);                       // 書き込み信号セット

    work    |= 0x04;                        // E信号セット
    work    |= (value << 4);                // 書き込みデータを上位4bitに設定
    Wire.write(work);                       // 書き込みデータとE信号アソート

    work    ^= 0x04;                        // E信号をクリア
    Wire.write(work);                       // E信号をネゲート

    return(Wire.endTransmission(1));        // データ送信とストップコンディション完了

}

/*******************************************************************************
End of function LCD_nibble_write
*******************************************************************************/


